# C++ benchmarking

Benchmarking in C++ is the process of measuring and comparing the performance of code. It's crucial for optimizing applications, comparing algorithms, and making informed decisions about implementation choices. Proper benchmarking helps developers identify bottlenecks, validate optimizations, and ensure code efficiency.

## Key Characteristics

- Measures execution time, memory usage, and other performance metrics
- Requires careful setup to ensure accurate and reproducible results
- Often involves multiple runs to account for system variability
- Can be performed at various levels: from individual functions to entire applications
- Requires consideration of compiler optimizations and hardware factors
- Useful for comparing different implementations or algorithms


## Example 1: Basic Manual Timing

```cpp
#include <iostream>
#include <chrono>
#include <vector>
#include <algorithm>

// Function to be benchmarked
void sortVector(std::vector<int>& vec) {
    std::sort(vec.begin(), vec.end());
}

int main() {
    const int NUM_RUNS = 100;
    const int VECTOR_SIZE = 10000;

    std::vector<int> vec(VECTOR_SIZE);

    // Measure total time for multiple runs
    auto start = std::chrono::high_resolution_clock::now();

    for (int i = 0; i < NUM_RUNS; ++i) {
        // Reset vector for each run
        for (int j = 0; j < VECTOR_SIZE; ++j) {
            vec[j] = rand() % VECTOR_SIZE;
        }

        sortVector(vec);
    }

    auto end = std::chrono::high_resolution_clock::now();

    // Calculate average time
    auto duration = std::chrono::duration_cast<std::chrono::microseconds>(end - start);
    double averageTime = duration.count() / static_cast<double>(NUM_RUNS);

    std::cout << "Average time to sort " << VECTOR_SIZE << " integers: " 
              << averageTime << " microseconds" << std::endl;

    return 0;
}
```

### Explanation:
- This example demonstrates a basic manual benchmarking approach
- It measures the time taken to sort a vector of integers
- The benchmark is run multiple times to get an average performance
- `std::chrono` is used for high-precision timing
- The vector is reset with random values before each sorting operation

## Example 2: Using Google Benchmark Library

```cpp
#include <benchmark/benchmark.h>
#include <vector>
#include <algorithm>

static void BM_VectorSort(benchmark::State& state) {
    for (auto _ : state) {
        state.PauseTiming();
        std::vector<int> vec(state.range(0));
        for (int& i : vec) {
            i = rand() % state.range(0);
        }
        state.ResumeTiming();

        std::sort(vec.begin(), vec.end());
    }
}

BENCHMARK(BM_VectorSort)
    ->Range(8, 8<<10);

BENCHMARK_MAIN();
```

### Explanation:
- This example uses the Google Benchmark library for more sophisticated benchmarking
- The `BM_VectorSort` function is the benchmark
- `state.PauseTiming()` and `state.ResumeTiming()` are used to exclude setup time from the measurement
- The benchmark is run for different vector sizes using `Range(8, 8<<10)`
- Google Benchmark handles multiple runs and statistical analysis automatically

To compile and run (assuming Google Benchmark is installed):
```
g++ -std=c++11 -isystem benchmark/include -Lbenchmark/build/src -lbenchmark -lpthread benchmark_example.cpp -o benchmark_example
./benchmark_example
```

## Example 3: Memory Usage Benchmarking

```cpp
#include <iostream>
#include <vector>
#include <chrono>
#include <memory>
#include <cstdio>

#ifdef __unix__
#include <unistd.h>
#include <sys/resource.h>
#elif defined(_WIN32)
#include <windows.h>
#include <psapi.h>
#endif

// A simple class to allocate memory
class MemoryUser {
    std::unique_ptr<char[]> buffer;
public:
    MemoryUser(size_t size) : buffer(new char[size]) {}
};

// Function to measure memory usage
size_t getCurrentRSS() {
#ifdef __unix__
    struct rusage rusage;
    if (getrusage(RUSAGE_SELF, &rusage) != 0) {
        return 0;
    }
    return (size_t)(rusage.ru_maxrss * 1024L);
#elif defined(_WIN32)
    PROCESS_MEMORY_COUNTERS_EX pmc;
    if (GetProcessMemoryInfo(GetCurrentProcess(), (PROCESS_MEMORY_COUNTERS*)&pmc, sizeof(pmc))) {
        return pmc.WorkingSetSize;
    }
    return 0;
#else
    // For other operating systems, you might need to implement a different method
    return 0;
#endif
}

int main() {
    const int NUM_OBJECTS = 1000000;
    const int OBJECT_SIZE = 1000;

    size_t initialMemory = getCurrentRSS();

    auto start = std::chrono::high_resolution_clock::now();

    std::vector<MemoryUser> objects;
    objects.reserve(NUM_OBJECTS);

    for (int i = 0; i < NUM_OBJECTS; ++i) {
        objects.emplace_back(OBJECT_SIZE);
    }

    auto end = std::chrono::high_resolution_clock::now();
    size_t finalMemory = getCurrentRSS();

    auto duration = std::chrono::duration_cast<std::chrono::milliseconds>(end - start);

    std::cout << "Time taken: " << duration.count() << " milliseconds" << std::endl;
    std::cout << "Memory used: " << (finalMemory - initialMemory) / (1024 * 1024) << " MB" << std::endl;

    return 0;
}
```

### Explanation:
- This example demonstrates benchmarking both execution time and memory usage
- The `MemoryUser` class is used to allocate a specific amount of memory
- `getCurrentRSS()` is a simplified function to get the current memory usage (works on Linux)
- The benchmark measures the time and memory needed to create a large number of `MemoryUser` objects
- Memory usage is calculated by comparing before and after the operation
- We've replaced the Linux-specific `/proc/self/statm` method with a more portable approach using `getrusage()`.
- We're now using `#ifdef __unix__` to cover both Linux and macOS (and other Unix-like systems).
- The `getCurrentRSS()` function now uses `getrusage()` on Unix-like systems:
    - `getrusage()` provides the maximum resident set size of the process.
    - The value is typically in kilobytes, so we multiply by 1024 to get bytes.
- The Windows implementation remains the same.
- We've removed the dependency on `_SC_PAGESIZE`, which was causing the error.

This example demonstrates how to create more portable benchmarking code that can work across different operating systems, which is crucial for cross-platform development and benchmarking.

## Example 4: Comparative Benchmarking

```cpp
#include <iostream>
#include <vector>
#include <algorithm>
#include <chrono>
#include <random>
#include <functional>

// Timing function
template<typename Func>
long long measureTime(Func func) {
    auto start = std::chrono::high_resolution_clock::now();
    func();
    auto end = std::chrono::high_resolution_clock::now();
    return std::chrono::duration_cast<std::chrono::microseconds>(end - start).count();
}

// Sorting algorithms to compare
void bubbleSort(std::vector<int>& vec) {
    for (size_t i = 0; i < vec.size(); ++i)
        for (size_t j = 0; j < vec.size() - 1; ++j)
            if (vec[j] > vec[j + 1])
                std::swap(vec[j], vec[j + 1]);
}

void quickSort(std::vector<int>& vec) {
    std::sort(vec.begin(), vec.end());
}

int main() {
    const int NUM_RUNS = 10;
    const int VECTOR_SIZE = 10000;

    std::vector<int> original(VECTOR_SIZE);
    std::random_device rd;
    std::mt19937 gen(rd());
    std::uniform_int_distribution<> dis(1, VECTOR_SIZE);

    // Initialize vector with random numbers
    for (int& num : original) {
        num = dis(gen);
    }

    // Benchmark bubble sort
    long long bubbleSortTime = 0;
    for (int i = 0; i < NUM_RUNS; ++i) {
        std::vector<int> vec = original;
        bubbleSortTime += measureTime([&]() { bubbleSort(vec); });
    }

    // Benchmark quick sort
    long long quickSortTime = 0;
    for (int i = 0; i < NUM_RUNS; ++i) {
        std::vector<int> vec = original;
        quickSortTime += measureTime([&]() { quickSort(vec); });
    }

    std::cout << "Average Bubble Sort time: " << bubbleSortTime / NUM_RUNS << " microseconds" << std::endl;
    std::cout << "Average Quick Sort time: " << quickSortTime / NUM_RUNS << " microseconds" << std::endl;

    return 0;
}
```

### Explanation:
- This example compares the performance of two sorting algorithms: bubble sort and quick sort
- A template function `measureTime` is used for timing
- The same random vector is used for both algorithms to ensure fair comparison
- Multiple runs are performed to get average performance
- The results clearly show the performance difference between the two algorithms

## Additional Considerations

1. **Compiler Optimizations**: Be aware of compiler optimizations that might affect benchmarking results. Sometimes, you may need to disable certain optimizations for accurate measurements.

2. **Hardware Factors**: CPU frequency scaling, thermal throttling, and other hardware factors can affect benchmark results. Ensure consistent hardware state for reliable comparisons.

3. **Microbenchmarking Pitfalls**: Very small benchmarks might be affected by factors like cache effects or branch prediction. Be cautious when interpreting results of microbenchmarks.

4. **Statistical Analysis**: For more robust results, consider using statistical methods to analyze benchmark data, such as calculating standard deviation or using confidence intervals.

5. **Realistic Workloads**: Ensure that your benchmarks represent realistic use cases. Synthetic benchmarks might not always reflect real-world performance.

6. **Profiling Tools**: Consider using profiling tools in addition to benchmarking for a more comprehensive performance analysis.

## Summary

Benchmarking in C++ is an essential practice for performance optimization and comparison. Key points to remember:

- It involves measuring execution time, memory usage, and other performance metrics
- Proper setup and multiple runs are crucial for accurate results
- Tools like Google Benchmark can simplify the benchmarking process
- Consider both time and memory usage in benchmarks
- Comparative benchmarking helps in choosing between different implementations
- Be aware of factors like compiler optimizations and hardware state
- Use realistic workloads and consider statistical analysis for robust results

Effective benchmarking is a skill that develops with practice. It's an invaluable tool for C++ developers looking to optimize their code and make informed decisions about implementation choices. By understanding and applying proper benchmarking techniques, developers can ensure their applications perform efficiently and meet performance requirements.

## Related 

- What are the best practices for benchmarking in C++
- How can I measure the performance of a specific function in C++
- What tools are available for benchmarking in C++
- How do I compare the performance of different data structures in C++
- What are common pitfalls when benchmarking in C++
